@@ -150,7 +150,7 @@ var _ = Describe("Podman kill", func() {
 
 	It("podman kill --cidfile", func() {
 		tmpDir, err := os.MkdirTemp("", "")
-		Expect(err).To(BeNil())
+		Expect(err).ToNot(HaveOccurred())
 		tmpFile := tmpDir + "cid"
 		defer os.RemoveAll(tmpDir)
 
@@ -170,12 +170,12 @@ var _ = Describe("Podman kill", func() {
 
 	It("podman kill multiple --cidfile", func() {
 		tmpDir1, err := os.MkdirTemp("", "")
-		Expect(err).To(BeNil())
+		Expect(err).ToNot(HaveOccurred())
 		tmpFile1 := tmpDir1 + "cid"
 		defer os.RemoveAll(tmpDir1)
 
 		tmpDir2, err := os.MkdirTemp("", "")
-		Expect(err).To(BeNil())
+		Expect(err).ToNot(HaveOccurred())
 		tmpFile2 := tmpDir2 + "cid"
 		defer os.RemoveAll(tmpDir2)
 
